/**
 * Infrared remote control signal Rx
 *  
 * MPLAB X IDE v5.10 with XC8 Ver2.05
 * (c)2020 Futoshi Sumikawa.
 */

#include "irrxsync.h"

uint8_t (*_irrxGetIR)(void);
void (*_irrxResetTimer)(void);
uint16_t (*_irrxReadTimer)(void);

#define _irrxIrInGetValue() _irrxGetIR()

uint8_t* _irrxData;

// Test 5key Panasonic LED ON
//uint8_t irFrameData[32] = {0x07, 0x40, 0x03, 0x83, 0x00, 0xFD, 0x00, 0x00,
//                           0x00, 0xCC, 0x02, 0x8B, 0x79, 0x00, 0x00, 0x28,
//                           0x34, 0x4A, 0x90, 0xB4, 0x24, 0x00, 0x00, 0x00,
//                           0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};


// IR frame data format.
//
//  00  01  02  03  04  05  06  07  08  09  0A  0B  0C  0D  0E  0F
// [LeadH ][LeadL ][DataHS][DataHL][DataLS][DataLL][Trail ][Number]
//  10  11  12  13  14  15  16  17  18  19  1A  1B  1C  1D  1E  1F
// [D0][D1][D2][D3][D4][D5][D6][D7][D8][D9][DA][DB][DC][DD][DE][DF]
//
// LeadH: LeaderH x 2us
// LeadL: LeaderL x 2us
// DataHS: HShort x 2us
// DataHL: HLong  x 2us
// DataLS: LShort x 2us
// DataLL: LLong  x 2us
// Trail: Trailer x 2us
// Number : Number of data bits
// D0-DF: Data 128bit
//


// AEHA Format.
// |~~~~~~~~|____|~|_|~|___...|~|____________|~~~~...
//  Leader        0   1        Trailer
//  8T       4T   T T T 3T     (>=8ms)130ms->|Leader
// T = 350 - 500 (425us typ.)
// Variable length data (48bit typ.)


// NEC Format.
// |~~~~~~~~~~~~~~~~|________|~|_|~|___...|~|____________|~~~~...
//  Leader                    0   1        Trailer    
//  16T              8T       T T T 3T            108ms->|Leader
// T = 562us
// Data 32bit


// Sony Format.
// |~~~~~~|_|~|_|~~|...____________|~~~~...
//  Leader 0   1             45ms->|Leader
//  4T     T T T 2T
// T = 600us
// Data 12bit or 15bit or 20bit

// Timeout 100ms (2us x 50000)
#define IRRX_CYCLE_TIMEOUT 50000
// 8ms (2us x 4000)
#define IRRX_CYCLE_8ms 4000 

#define IRRX_H 0
#define IRRX_L 1

uint16_t _irrxDataTrailer = 0;

uint8_t _irrxStartLeader = 0;

uint16_t _irrxDataLeaderH = 0;
uint16_t _irrxDataLeaderL = 0;

uint16_t _irrxDataHShort = 0;
uint16_t _irrxDataLShort = 0;

uint16_t _irrxDataHShortSum = 0;
uint16_t _irrxDataHLongSum = 0;
uint16_t _irrxDataLShortSum = 0;
uint16_t _irrxDataLLongSum = 0;

uint8_t _irrxDataHShortNumber = 0;
uint8_t _irrxDataHLongNumber = 0;
uint8_t _irrxDataLShortNumber = 0;
uint8_t _irrxDataLLongNumber = 0;

uint8_t _irrxDataHNumber = 0;
uint8_t _irrxDataLNumber = 0;

uint8_t _irrxComplete = 0;
uint8_t _irrxBusy = 0;


void _irrxInitialize() {

    _irrxDataLeaderH = 0;
    _irrxDataLeaderL = 0;

    _irrxDataHShort = 0;
    _irrxDataLShort = 0;

    _irrxDataHShortSum = 0;
    _irrxDataHLongSum = 0;
    _irrxDataLShortSum = 0;
    _irrxDataLLongSum = 0;

    _irrxDataHShortNumber = 0;
    _irrxDataHLongNumber = 0;
    _irrxDataLShortNumber = 0;
    _irrxDataLLongNumber = 0;

    _irrxDataHNumber = 0;
    _irrxDataLNumber = 0;

    for (uint8_t i = 0; i < 32; i++) {
        _irrxData[i] = 0;
    }

    _irrxComplete = 0;
    _irrxBusy = 0;
}

// a > b : 1
// a = b : 0
// a < b : 0xff
uint8_t _irrxCompare(uint16_t a, uint16_t b) {
    uint16_t harf = 0;
    if (a > b) {
        harf = b >> 1;
        if (a - b > harf) {
            return 1;
        }
    } else {
        harf = a >> 1;
        if (b - a > harf) {
            return 0xff;
        }
    }
    return 0;
}

void _irrxAddDataHShort(uint16_t value) {
    _irrxDataHShortSum += value;
    _irrxDataHShortNumber++;
}
void _irrxAddDataHLong(uint16_t value) {
    _irrxDataHLongSum += value;
    _irrxDataHLongNumber++;
}
void _irrxAddDataLShort(uint16_t value) {
    _irrxDataLShortSum += value;
    _irrxDataLShortNumber++;
}
void _irrxAddDataLLong(uint16_t value) {
    _irrxDataLLongSum += value;
    _irrxDataLLongNumber++;
}


void _irrxSetDataL(uint8_t number, uint8_t data) {
    
    uint8_t tableIndex = (number & 0b01111111) >> 3;
    uint8_t bitIndex = 7 - (number & 0b00000111);
    uint8_t mask = 0x01 << bitIndex;
    
    uint8_t dataByte = _irrxData[16 + tableIndex];
    dataByte = (dataByte & (0xff ^ mask)) | (data << bitIndex);
    _irrxData[16 + tableIndex] = dataByte;
    
}


void _irrxSetDataH(uint8_t number, uint8_t data) {
    
    uint8_t tableIndex = (number & 0b01111111) >> 3;
    uint8_t bitIndex = 7 - (number & 0b00000111);
    uint8_t mask = 0x01 << bitIndex;
    
    uint8_t dataByte = _irrxData[tableIndex];
    dataByte = (dataByte & (0xff ^ mask)) | (data << bitIndex);
    _irrxData[tableIndex] = dataByte;
    
}

void _irrxEdge(uint8_t _irrxPinStatus, uint16_t _irrxCycleNumber) {

    uint8_t data = 0;
    
    if (_irrxPinStatus == IRRX_L) {
        
        // Positive edge.
        if (_irrxDataLeaderH == 0) {
            if (_irrxCycleNumber > IRRX_CYCLE_8ms) {
                _irrxStartLeader = 1;
            } else {
                _irrxStartLeader = 0;
            }
        } else if (_irrxDataLeaderL == 0) {
            // Leader L
            _irrxDataLeaderL = _irrxCycleNumber;
        } else if (_irrxStartLeader == 0) {

            // Trailer
            if (_irrxCycleNumber > IRRX_CYCLE_8ms) {
                _irrxStartLeader = 1;
                _irrxDataTrailer = _irrxCycleNumber;
                
                if (_irrxDataHLongNumber == 0 || _irrxDataLLongNumber == 0) {
                    _irrxComplete = 1;
                }
                return;
            }
            
            // Data L
            if (_irrxDataLShort == 0) {
                // Data L short.
                _irrxDataLShort = _irrxCycleNumber;
                _irrxAddDataLShort(_irrxCycleNumber);
            } else {
                uint8_t c = _irrxCompare(_irrxDataLShort, _irrxCycleNumber);
                if ( c == 0) {
                    // Data L short.
                    _irrxAddDataLShort(_irrxCycleNumber);
                } else if (c == 1) {
                    // Swap
                    _irrxDataLLongSum = _irrxDataLShortSum;
                    _irrxDataLLongNumber = _irrxDataLShortNumber;

//                    for (uint8_t i = 0; i < 16; i++) {
//                        _irrxData[16 + i]  = _irrxData[16 + i] ^ 0xff;
//                    }
                        // 2 bytes are processed assuming that the beginning is not 0xffff.
                        _irrxData[16]  = _irrxData[16] ^ 0xff;
                        _irrxData[17]  = _irrxData[17] ^ 0xff;
       
                    // Data L short.
                    _irrxDataLShort = _irrxCycleNumber;
                    _irrxAddDataLShort(_irrxCycleNumber);
                } else {
                    // Data L long.
                    data = 1;
                   _irrxAddDataLLong(_irrxCycleNumber);
                }
            }
            _irrxSetDataL(_irrxDataLNumber, data);
            _irrxDataLNumber++;
        }
    } else {
        
        // Negative edge.
        
        if (_irrxDataLeaderL == 0) {
            if (_irrxStartLeader == 1) {
                // Leader H
                _irrxDataLeaderH = _irrxCycleNumber;
                _irrxStartLeader = 0;
            }
        } else if (_irrxStartLeader == 0) {
            
            // Data H
            if (_irrxDataHShort == 0) {
                // Data H short.
                _irrxDataHShort = _irrxCycleNumber;
                _irrxAddDataHShort(_irrxCycleNumber);
            } else {
                uint8_t c = _irrxCompare(_irrxDataHShort, _irrxCycleNumber);
                if ( c == 0) {
                    // Data H short.
                    _irrxAddDataHShort(_irrxCycleNumber);
                } else if (c == 1) {
                    // Swap
                    _irrxDataHLongSum = _irrxDataHShortSum;
                    _irrxDataHLongNumber = _irrxDataHShortNumber;
                    
//                    for (uint8_t i = 0; i < 16; i++) {
//                        _irrxData[i]  = _irrxData[i] ^ 0xff;
//                    }
                        // 2 bytes are processed assuming that the beginning is not 0xffff.
                        _irrxData[0]  = _irrxData[0] ^ 0xff;
                        _irrxData[1]  = _irrxData[1] ^ 0xff;

                    // Data H short.
                    _irrxDataHShort = _irrxCycleNumber;
                    _irrxAddDataHShort(_irrxCycleNumber);
                } else {
                    // Data H long.
                    data = 1;
                   _irrxAddDataHLong(_irrxCycleNumber);
                }
            }
            _irrxSetDataH(_irrxDataHNumber, data);
            _irrxDataHNumber++;
        }
    }
    
}
#define IR_HARDWARE_OFFSET 48
void _irrxCrateFrameData(void) {

    if (_irrxDataHLongNumber == 0) {
        // AEHA
        // Data number
        _irrxData[14] = 0;
        _irrxData[15] = _irrxDataLNumber;    
    } else {
        // SONY
        // Data bits
        for (uint8_t i = 0; i < 16; i++) {
            _irrxData[16 + i] = _irrxData[i];
        }
        // Data number
        _irrxData[14] = 0;
        _irrxData[15] = _irrxDataHNumber;        
    }
    
    // Leader H (x2us)
    _irrxDataLeaderH -= IR_HARDWARE_OFFSET;
    _irrxData[0] = (_irrxDataLeaderH & 0xff00) >> 8;
    _irrxData[1] = _irrxDataLeaderH;
    // Leader L (x2us)
    _irrxDataLeaderL += IR_HARDWARE_OFFSET;
    _irrxData[2] = (_irrxDataLeaderL & 0xff00) >> 8;
    _irrxData[3] = _irrxDataLeaderL;
    // Data H short (x2us)
    _irrxDataHShortSum = _irrxDataHShortSum / _irrxDataHShortNumber - IR_HARDWARE_OFFSET;
    _irrxData[4] = (_irrxDataHShortSum & 0xff00) >> 8;
    _irrxData[5] = _irrxDataHShortSum;
    // Data H long (x2us)
    if (_irrxDataHLongSum != 0) {
        _irrxDataHLongSum = _irrxDataHLongSum / _irrxDataHLongNumber - IR_HARDWARE_OFFSET;
    }
    _irrxData[6] = (_irrxDataHLongSum & 0xff00) >> 8;
    _irrxData[7] = _irrxDataHLongSum;
    // Data L short (x2us)
    _irrxDataLShortSum = _irrxDataLShortSum / _irrxDataLShortNumber + IR_HARDWARE_OFFSET;
    _irrxData[8] = (_irrxDataLShortSum & 0xff00) >> 8;
    _irrxData[9] = _irrxDataLShortSum;
    // Data L long (x2us)
    if (_irrxDataLLongSum != 0) {
        _irrxDataLLongSum = _irrxDataLLongSum / _irrxDataLLongNumber + IR_HARDWARE_OFFSET;
    }
    _irrxData[10] = (_irrxDataLLongSum & 0xff00) >> 8;
    _irrxData[11] = _irrxDataLLongSum;
    // Trailer (x2us)
    _irrxDataTrailer += IR_HARDWARE_OFFSET;
    _irrxData[12] = (_irrxDataTrailer & 0xff00) >> 8;
    _irrxData[13] = _irrxDataTrailer;

    
}

uint8_t _irrxRecive() {
    
    while (_irrxIrInGetValue() == IRRX_H) {
        // Invalid
    }
    
    uint8_t irrxPin = IRRX_L;
    uint16_t offsetCycle = 0;
    _irrxResetTimer();
    _irrxBusy = 1;

    uint8_t GIEBitValue = INTCONbits.GIE;
    
    while (_irrxBusy && _irrxComplete == 0) {
        uint16_t cycle = _irrxReadTimer();
        if(_irrxIrInGetValue() == IRRX_H) {
            // Positive
            
            // Disable INTs
            INTCONbits.GIE = 0;
            
            if (irrxPin == IRRX_L) {
                // Positive edge
                _irrxEdge(IRRX_L, cycle - offsetCycle);
                if ((offsetCycle == 0 || _irrxComplete) && cycle > IRRX_CYCLE_8ms) {
                    // Start leader
                    _irrxResetTimer();
                } else {
                    offsetCycle = cycle;
                }
            }
            irrxPin = IRRX_H;
            
        } else {
            // Negative
            if (irrxPin == IRRX_H) {
                // Negative edge
                _irrxEdge(IRRX_H, cycle - offsetCycle);
                offsetCycle = cycle;
            } 
            irrxPin = IRRX_L;

            if (cycle - offsetCycle > IRRX_CYCLE_TIMEOUT) {
                // Timeout
                _irrxBusy = 0;
            }
        }
    }
    
    INTCONbits.GIE = GIEBitValue;
    
    if (_irrxComplete) {
        _irrxCrateFrameData();
        return 1;
    }
    return 0;
}




void irrxBegin(uint8_t* irFrameData, uint8_t (* getIR)(void), void (* resetTimer)(void), uint16_t (* readTimer)(void)) {
    _irrxData = irFrameData;
    _irrxGetIR = getIR;
    _irrxResetTimer = resetTimer;
    _irrxReadTimer = readTimer;
}


uint8_t irrxIsReceiveComplete() {
    _irrxInitialize();
    return _irrxRecive();
}

/**
 End of File
*/
