/**
 * ST7032
 * 
 * (c)2020 Futoshi Sumikawa.
 * 
 * 
 * 
 * 
 *  MSSP1
 *    -Hardware Settings
 *      Serial Protocol: I2C
 *      Mode: Master
 *
 * 
 */
#include "st7032.h"

#define _lcdCommand(c)     do { I2C1_Write1ByteRegister(ST7032_ADDRESS, 0x00, c); __delay_us(27); } while(0)

/**
 * Clear display.
 */
void lcdClear(void) {
    // Clear display.
    _lcdCommand(0x01);
    // Wait 1.08ms
    __delay_us(1100);
    // Return home.
    _lcdCommand(0x02);
    // Wait 1.08ms
    __delay_us(1100);
}
/**
 * Set cursor.
 * @param col
 * @param row
 */
void lcdSetCursor(uint8_t col, uint8_t row) {
    const uint8_t rowOffset[] = { 0x00, 0x40 };
    // Set DDRAM Address : 00H-27H,40H-67H
    _lcdCommand(0x80 | (col + rowOffset[row]));
}

/**
 * Put character.
 * @param c character
 */
void lcdPutc(uint8_t c){ 
    I2C1_Write1ByteRegister(ST7032_ADDRESS, 0x40, c);
    __delay_us(27);
}

void lcdPuts(const char * s) {
    uint8_t len = 0;
    for (len = 0; s[len] != '\0'; ++len);
    for (uint8_t i = 0; i < len; i++) {
        I2C1_Write1ByteRegister(ST7032_ADDRESS, 0x40, s[i]);
        __delay_us(27);
    }
}

/**
 * Set contrast.
 * @param contrast 0 - 63
 */
void lcdSetContrast(uint8_t contrast) {
    // enable extended command settings
    _lcdCommand(0x39);
    // lower 4 bits of contrast
    _lcdCommand(0x70 | (contrast & 0x0F));
    // upper 2 bits of contrast
    _lcdCommand(0x54 | ((contrast & 0x30) >> 4));
    // disable extended command settings
    _lcdCommand(0x38);
}

/**
 * Register char.
 * @param address 0 - 5
 * @param data 7byte
 */
void lcdRegisterChar(uint8_t address, uint8_t * data) {

    I2C1_Write1ByteRegister(ST7032_ADDRESS, 0x00, (uint8_t)(0x40 | address << 3));
    __delay_us(27);

    uint8_t buff[9];
    buff[0] = 0x40;
    for (uint8_t i = 0; i < 8; i++) {
        buff[i + 1]  = data[i];
    }
    I2C1_WriteNBytes(ST7032_ADDRESS, buff, 9);
}


/**
 * Initialize ST7032
 */
void lcdInitialize(uint8_t contrast) {
    __delay_ms(40);
    // function set: 8 data lines, 2 lines for display, 5x8 dots for font
    _lcdCommand(0x38);
    // function set: Enable extended command setting
    _lcdCommand(0x39);
    // Internal OSC frequency: Select bias and adjust internal OSC frequency
    _lcdCommand(0x14);
    // lower 4 bits of contrast
    _lcdCommand(0x70 | (contrast & 0x0F));
    // upper 2 bits of contrast
    _lcdCommand(0x54 | ((contrast & 0x30) >> 4));
    // follower control: Turn on follower circuit and adjust gain
    _lcdCommand(0x6C);
    __delay_ms(200);
    // function set: Disable setting of extended command
    _lcdCommand(0x38);
    // display control: Screen display ON, cursor display OFF, cursor blinking OFF
    _lcdCommand(0x0C);
    
}


/**
 * Begin ST7032
 */
void lcdBegin(void) {
    lcdInitialize(32);
    lcdClear();    
}
